# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = (
    "MigrationSessionImpl",
    "MigrationSession",
    "MongoSessionContextT",
    "MongoSessionContext",
    "TransactionT",
    "TransactionContext",
    "requires_auditlog",
)

import abc
import contextlib
import datetime
import functools
import logging
import typing
import uuid

import bson
import typing_extensions

from mongorunway import mongo
from mongorunway import util
from mongorunway.application import transactions
from mongorunway.domain import migration as domain_migration
from mongorunway.domain import migration_auditlog_entry as domain_auditlog_entry

if typing.TYPE_CHECKING:
    from mongorunway.application import applications
    from mongorunway.application import config
    from mongorunway.application.ports import filename_strategy as filename_strategy_port
    from mongorunway.domain import migration_event as domain_event

try:
    _P = typing.ParamSpec("_P")
except AttributeError:
    _P = typing_extensions.ParamSpec("_P")

_T = typing.TypeVar("_T")

TransactionT = typing.TypeVar("TransactionT", bound=transactions.MigrationTransaction)
MongoSessionContextT = typing.TypeVar("MongoSessionContextT", bound="MongoSessionContext")

_LOGGER: typing.Final[logging.Logger] = logging.getLogger("mongorunway.session")


def requires_auditlog(meth: typing.Callable[_P, _T]) -> typing.Callable[_P, _T]:
    @functools.wraps(meth)
    def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> _T:
        self: "MigrationSession" = typing.cast("MigrationSession", args[0])
        if not self.uses_auditlog:
            raise ValueError("Auditlog is not enabled on this session.")

        return meth(*args, **kwargs)

    return typing.cast(typing.Callable[_P, _T], wrapper)


class TransactionContext(typing.Generic[TransactionT]):
    def __init__(
        self,
        migration_session: MigrationSession,
        transaction_type: typing.Type[TransactionT],
        migration: domain_migration.Migration,
    ) -> None:
        self._transaction = transaction_type.create(migration_session, migration)
        self._migration = migration
        self._migration_session = migration_session

    def __enter__(self) -> TransactionT:
        return self._transaction

    def __exit__(self, *args: typing.Any) -> None:
        if self._migration_session.uses_auditlog:
            entry = self._build_auditlog_entry()

            if self._transaction.is_failed():
                assert self._transaction.exc_val is not None  # For type checkers only
                entry.with_error(self._transaction.exc_val)

            self._migration_session.log_audit_entry(entry)

    def _build_auditlog_entry(self) -> domain_auditlog_entry.MigrationAuditlogEntry:
        entry = domain_auditlog_entry.MigrationAuditlogEntry(
            date_fmt=self._migration_session.session_date_format,
            session_id=self._migration_session.session_id,
            transaction_name=type(self._transaction).__name__,
            migration_read_model=domain_migration.MigrationReadModel.from_migration(
                self._migration,
            ),
        )

        if self._migration_session.uses_unique_timezone:
            entry.with_timezone(self._migration_session.session_timezone)

        return entry


class MongoSessionContext:
    # incapsulate
    def __init__(self, session: mongo.ClientSession) -> None:
        self.__session = session

    def __enter__(self: MongoSessionContextT) -> MongoSessionContextT:
        return self

    def __exit__(self, *args: typing.Any) -> None:
        self.__session.end_session()

    @property
    def mongodb_session_id(self) -> bson.binary.Binary:
        return typing.cast(
            bson.binary.Binary,
            self.__session.session_id["id"],
        )

    @property
    def has_ended(self) -> bool:
        return self.__session.has_ended

    def start_transaction(
        self,
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> contextlib.AbstractContextManager[typing.Any]:
        return self.__session.start_transaction(*args, **kwargs)

    def commit_transaction(self) -> None:
        return self.__session.commit_transaction()

    def abort_transaction(self) -> None:
        return self.__session.abort_transaction()


class MigrationSession(abc.ABC):
    __slots__ = ()

    @property
    @abc.abstractmethod
    def session_id(self) -> bson.binary.Binary:
        ...

    @property
    @abc.abstractmethod
    def session_timezone(self) -> str:
        ...

    @property
    @abc.abstractmethod
    def session_date_format(self) -> str:
        ...

    @property
    @abc.abstractmethod
    def session_client(self) -> mongo.Client:
        ...

    @property
    @abc.abstractmethod
    def session_database(self) -> mongo.Database:
        ...

    @property
    @abc.abstractmethod
    def session_name(self) -> str:
        ...

    @property
    @abc.abstractmethod
    def session_subscribed_events(
        self,
    ) -> typing.Mapping[
        typing.Type[domain_event.MigrationEvent],
        typing.Sequence[domain_event.EventHandlerProxyOr[domain_event.EventHandler]],
    ]:
        ...

    @property
    @abc.abstractmethod
    def session_auditlog_limit(self) -> typing.Optional[int]:
        ...

    @property
    @abc.abstractmethod
    def session_config_dir(self) -> str:
        ...

    @property
    @abc.abstractmethod
    def session_scripts_dir(self) -> str:
        ...

    @property
    @abc.abstractmethod
    def session_file_naming_strategy(self) -> filename_strategy_port.FilenameStrategy:
        ...

    @property
    @abc.abstractmethod
    def uses_strict_file_naming(self) -> bool:
        ...

    @property
    @abc.abstractmethod
    def uses_schema_validation(self) -> bool:
        ...

    @property
    @abc.abstractmethod
    def uses_indexing(self) -> bool:
        ...

    @property
    @abc.abstractmethod
    def uses_logging(self) -> bool:
        ...

    @property
    @abc.abstractmethod
    def uses_unique_timezone(self) -> bool:
        ...

    @property
    @abc.abstractmethod
    def uses_auditlog(self) -> bool:
        ...

    @property
    @abc.abstractmethod
    def raises_on_transaction_failure(self) -> bool:
        ...

    @abc.abstractmethod
    def history(
        self,
        start: typing.Optional[datetime.datetime] = None,
        end: typing.Optional[datetime.datetime] = None,
        limit: typing.Optional[int] = None,
        ascending_date: bool = True,
    ) -> typing.Iterator[domain_auditlog_entry.MigrationAuditlogEntry]:
        ...

    @abc.abstractmethod
    def has_migration(self, item: typing.Any, /) -> bool:
        ...

    @abc.abstractmethod
    def has_migration_with_version(self, migration_version: int, /) -> bool:
        ...

    @abc.abstractmethod
    def has_migrations(self) -> bool:
        ...

    @abc.abstractmethod
    def get_migration_model_by_version(
        self,
        migration_version: int,
    ) -> typing.Optional[domain_migration.MigrationReadModel]:
        ...

    @abc.abstractmethod
    def get_migration_model_by_flag(
        self,
        *,
        is_applied: bool,
    ) -> typing.Optional[domain_migration.MigrationReadModel]:
        ...

    @abc.abstractmethod
    def get_all_migration_models(
        self,
        *,
        ascending_id: bool = True,
    ) -> typing.MutableSequence[domain_migration.MigrationReadModel]:
        ...

    @abc.abstractmethod
    def get_migration_models_by_flag(
        self, *, is_applied: bool
    ) -> typing.MutableSequence[domain_migration.MigrationReadModel]:
        ...

    @abc.abstractmethod
    def append_migration(self, migration: domain_migration.Migration, /) -> None:
        ...

    @abc.abstractmethod
    def remove_migration(self, migration_version: int, /) -> None:
        ...

    @abc.abstractmethod
    def get_current_version(self) -> typing.Optional[int]:
        ...

    @abc.abstractmethod
    def set_applied_flag(self, migration: domain_migration.Migration, is_applied: bool) -> None:
        ...

    @abc.abstractmethod
    def log_audit_entry(
        self, migration_auditlog_entry: domain_auditlog_entry.MigrationAuditlogEntry, /
    ) -> None:
        ...

    @abc.abstractmethod
    def begin_mongo_session(self) -> MongoSessionContext:
        ...

    @abc.abstractmethod
    def begin_transaction(
        self,
        transaction_type: typing.Type[TransactionT],
        migration: domain_migration.Migration,
    ) -> TransactionContext[TransactionT]:
        ...


class MigrationSessionImpl(MigrationSession):
    def __init__(
        self,
        app: applications.MigrationApp,
        configuration: config.Config,
    ) -> None:
        self._config = configuration
        self._application = app

        self._repository = configuration.application.app_repository
        self._auditlog_journal = configuration.application.app_auditlog_journal

    @property
    def session_id(self) -> bson.binary.Binary:
        return bson.binary.Binary(uuid.uuid4().bytes, bson.UUID_SUBTYPE)  # like in pymongo session

    @property
    def session_timezone(self) -> str:
        return self._config.application.app_timezone

    @property
    def session_date_format(self) -> str:
        return self._config.application.app_date_format

    @property
    def session_client(self) -> mongo.Client:
        return self._config.application.app_client

    @property
    def session_database(self) -> mongo.Database:
        return self._config.application.app_database

    @property
    def session_name(self) -> str:
        return self._config.application.app_name

    @property
    def session_subscribed_events(
        self,
    ) -> typing.Mapping[
        typing.Type[domain_event.MigrationEvent],
        typing.Sequence[domain_event.EventHandlerProxyOr[domain_event.EventHandler]],
    ]:
        return self._config.application.app_subscribed_events

    @property
    def session_auditlog_limit(self) -> typing.Optional[int]:
        return self._config.application.app_auditlog_limit

    @property
    def session_config_dir(self) -> str:
        return self._config.filesystem.config_dir

    @property
    def session_scripts_dir(self) -> str:
        return self._config.filesystem.scripts_dir

    @property
    def session_file_naming_strategy(self) -> filename_strategy_port.FilenameStrategy:
        return self._config.filesystem.filename_strategy

    @property
    def uses_strict_file_naming(self) -> bool:
        return self._config.filesystem.use_filename_strategy

    @property
    def uses_schema_validation(self) -> bool:
        return self._config.application.use_schema_validation

    @property
    def uses_indexing(self) -> bool:
        return self._config.application.use_indexing

    @property
    def uses_logging(self) -> bool:
        return self._config.application.is_logged

    @property
    def uses_unique_timezone(self) -> bool:
        return self._config.application.has_unique_timezone

    @property
    def uses_auditlog(self) -> bool:
        return self._config.application.is_auditlog_enabled

    @property
    def raises_on_transaction_failure(self) -> bool:
        return self._config.application.raise_on_transaction_failure

    def has_migration(self, item: typing.Any, /) -> bool:
        return self._repository.has_migration(item)

    def has_migration_with_version(self, migration_version: int, /) -> bool:
        if not isinstance(migration_version, int):
            raise TypeError(f"Migration version must be instance of {int!r}.")

        return self._repository.has_migration_with_version(migration_version)

    def has_migrations(self) -> bool:
        return self._repository.has_migrations()

    def get_migration_model_by_version(
        self, migration_version: int
    ) -> typing.Optional[domain_migration.MigrationReadModel]:
        if not isinstance(migration_version, int):
            raise TypeError(f"Migration version must be instance of {int!r}.")

        return self._repository.acquire_migration_model_by_version(migration_version)

    def get_migration_model_by_flag(
        self,
        *,
        is_applied: bool,
    ) -> typing.Optional[domain_migration.MigrationReadModel]:
        return self._repository.acquire_migration_model_by_flag(is_applied)

    def get_all_migration_models(
        self,
        *,
        ascending_id: bool = True,
    ) -> typing.MutableSequence[domain_migration.MigrationReadModel]:
        return list(self._repository.acquire_all_migration_models(ascending_id=ascending_id))

    def get_migration_models_by_flag(
        self,
        *,
        is_applied: bool,
    ) -> typing.MutableSequence[domain_migration.MigrationReadModel]:
        return list(self._repository.acquire_migration_models_by_flag(is_applied=is_applied))

    def get_current_version(self) -> typing.Optional[int]:
        if (
            target := self._repository.acquire_migration_model_by_flag(is_applied=True)
        ) is not None:
            return typing.cast(
                typing.Optional[int],
                target.version,
            )

        return target

    def append_migration(self, migration: domain_migration.Migration, /) -> None:
        if not isinstance(migration, domain_migration.Migration):
            raise TypeError(f"Migration must be instance of {domain_migration.Migration!r}.")

        self._repository.append_migration(migration)

    def remove_migration(self, migration_version: int, /) -> None:
        if not isinstance(migration_version, int):
            raise TypeError(f"Migration version must be instance of {int!r}.")

        model = self._repository.acquire_migration_model_by_version(migration_version)
        if model is None:
            raise ValueError(f"Migration with version {migration_version} does not exist.")

        self._repository.remove_migration(migration_version)

    def set_applied_flag(self, migration: domain_migration.Migration, is_applied: bool) -> None:
        if not isinstance(migration, domain_migration.Migration):
            raise TypeError(f"Migration must be instance of {domain_migration.Migration!r}.")

        self._repository.set_applied_flag(migration, is_applied)

    @requires_auditlog
    def log_audit_entry(self, entry: domain_auditlog_entry.MigrationAuditlogEntry) -> None:
        if not isinstance(entry, domain_auditlog_entry.MigrationAuditlogEntry):
            raise TypeError(
                f"Migration auditlog entry must be instance of "
                f"{domain_auditlog_entry.MigrationAuditlogEntry!r}."
            )

        assert self._auditlog_journal is not None  # For type checkers only
        self._auditlog_journal.append_entries([entry])

    @requires_auditlog
    def history(
        self,
        start: typing.Optional[datetime.datetime] = None,
        end: typing.Optional[datetime.datetime] = None,
        limit: typing.Optional[int] = None,
        ascending_date: bool = True,
    ) -> typing.Iterator[domain_auditlog_entry.MigrationAuditlogEntry]:
        assert self._auditlog_journal is not None  # For type checkers only
        yield from self._auditlog_journal.history(
            start=start,
            end=end,
            limit=limit,
            ascending_date=ascending_date,
        )

    def begin_mongo_session(self) -> MongoSessionContext:
        ctx = MongoSessionContext(self.session_client.start_session())
        _LOGGER.info(
            "Mongorunway MongoDB context successfully initialized " "with MongoDB session id (%s)",
            util.hexlify(ctx.mongodb_session_id),
        )
        return ctx

    def begin_transaction(
        self,
        transaction_type: typing.Type[TransactionT],
        migration: domain_migration.Migration,
    ) -> TransactionContext[TransactionT]:
        ctx = TransactionContext(self, transaction_type, migration)
        _LOGGER.info(
            "Mongorunway transaction context successfully initialized "
            "with Mongorunway session id (%s)",
            util.hexlify(self.session_id),
        )
        return ctx



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = (
    "migration_file_template",
    "MigrationService",
)

import os
import string
import typing

from mongorunway import util
from mongorunway.application import config
from mongorunway.application.services import checksum_service
from mongorunway.domain import migration as domain_migration
from mongorunway.domain import migration_business_module as domain_module

if typing.TYPE_CHECKING:
    from mongorunway.application import session

migration_file_template = string.Template(
    """\
from __future__ import annotations

import typing

import mongorunway

# Required, used by Mongorunway.
version = $version


@mongorunway.migration
def upgrade() -> typing.Sequence[mongorunway.MigrationCommand]:
    return $upgrade_commands


@mongorunway.migration
def downgrade() -> typing.Sequence[mongorunway.MigrationCommand]:
    return $downgrade_commands
"""
)


class MigrationService:
    def __init__(self, app_session: session.MigrationSession) -> None:
        self._session = app_session

    def get_migration(
        self,
        migration_name: str,
        migration_version: int,
    ) -> domain_migration.Migration:
        if self._session.uses_strict_file_naming:
            strategy = self._session.session_file_naming_strategy
            migration_name = strategy.transform_migration_filename(
                migration_name,
                migration_version,
            )

        module = util.get_module(self._session.session_scripts_dir, migration_name)
        migration_module = domain_module.MigrationBusinessModule(module)

        model = self._session.get_migration_model_by_version(module.version)

        migration = domain_migration.Migration(
            name=migration_module.get_name(),
            version=module.version,
            description=migration_module.description,
            checksum=checksum_service.calculate_migration_checksum(migration_module),
            downgrade_process=migration_module.downgrade_process,
            upgrade_process=migration_module.upgrade_process,
            is_applied=False if model is None else model.is_applied,
        )

        return migration

    @typing.no_type_check
    def get_migrations(self) -> typing.Sequence[domain_migration.Migration]:
        filename_strategy = self._session.session_file_naming_strategy
        directory = self._session.session_scripts_dir

        if self._session.uses_strict_file_naming:
            # All migrations are in the correct order by name.
            return [
                self.get_migration(
                    filename_strategy.transform_migration_filename(migration_name, position),
                    position,
                )
                for position, migration_name in enumerate(
                    sorted(os.listdir(directory)),
                    config.VERSIONING_STARTS_FROM,
                )
                if util.is_valid_filename(directory, migration_name)
            ]

        else:
            migrations: typing.Dict[int, domain_migration.Migration] = {}
            for migration_name in sorted(os.listdir(directory)):
                if not util.is_valid_filename(directory, migration_name):
                    continue

                module = util.get_module(directory, migration_name)
                try:
                    migration_version = module.version
                except AttributeError:
                    raise ImportError(
                        f"Migration {migration_name} in non-strict mode must have "
                        f"'version' variable."
                    )

                migrations[migration_version] = self.get_migration(
                    migration_name,
                    migration_version,
                )

            if (start := config.VERSIONING_STARTS_FROM) not in migrations:
                # ...
                raise ValueError(f"Versioning starts from {start}.")

            return [migrations[key] for key in sorted(migrations.keys())]

    def create_migration_file_template(
        self,
        migration_filename: str,
        migration_version: typing.Optional[int] = None,
    ) -> None:
        migrations = self.get_migrations()
        if migration_version is None:
            migration_version = len(migrations) + 1

        if self._session.has_migration_with_version(migration_version):
            raise ValueError(f"Migration with version {migration_version} already exist.")

        if abs(len(migrations) - migration_version) > 1:
            raise ValueError(
                f"Versions of migrations must be consistent: the next version "
                f"must be {len(migrations) + 1!r}, but {migration_version!r} received."
            )

        filename_strategy = self._session.session_file_naming_strategy
        if self._session.uses_strict_file_naming:
            migration_filename = filename_strategy.transform_migration_filename(
                migration_filename,
                migration_version,
            )

            if not migration_filename.endswith(".py"):
                migration_filename += ".py"

        with open(
            os.path.join(
                self._session.session_scripts_dir,
                migration_filename,
            ),
            "w",
        ) as file:
            file.write(
                migration_file_template.safe_substitute(
                    version=migration_version,
                    upgrade_commands=[],
                    downgrade_commands=[],
                )
            )

        return None



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = ("MigrationBusinessModule",)

import types
import typing

from mongorunway.domain import migration as domain_migration


class MigrationBusinessModule:
    __slots__: typing.Sequence[str] = (
        "_module",
        "_upgrade_process",
        "_downgrade_process",
    )

    def __init__(self, module: types.ModuleType, /) -> None:
        self._module = module
        self._upgrade_process = self._get_business_process("upgrade")
        self._downgrade_process = self._get_business_process("downgrade")

    @property
    def location(self) -> str:
        return self._module.__file__ or ""

    @property
    def description(self) -> str:
        return self._module.__doc__ or ""

    @property
    def version(self) -> int:
        return typing.cast(int, self._module.version)

    @property
    def upgrade_process(self) -> domain_migration.MigrationProcess:
        return self._upgrade_process

    @property
    def downgrade_process(self) -> domain_migration.MigrationProcess:
        return self._downgrade_process

    def get_name(self) -> str:
        *_, migration_name = self._module.__name__.split(".")
        return migration_name.strip()

    def _get_business_process(self, process_name: str, /) -> domain_migration.MigrationProcess:
        process = getattr(self._module, process_name, None)
        if process is None:
            raise ValueError(
                f"Can't find {process_name!r} process in {self.get_name()!r} migration."
            )

        return process



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = ("MigrationModelRepository",)

import abc
import typing

if typing.TYPE_CHECKING:
    from mongorunway.domain import migration as domain_migration


class MigrationModelRepository(abc.ABC):
    __slots__: typing.Sequence[str] = ()

    @abc.abstractmethod
    def __len__(self) -> int:
        ...

    @abc.abstractmethod
    def __contains__(self, item: typing.Any, /) -> bool:
        ...

    @abc.abstractmethod
    def has_migration(self, item: typing.Any, /) -> bool:
        ...

    @abc.abstractmethod
    def has_migration_with_version(self, migration_version: int, /) -> bool:
        ...

    @abc.abstractmethod
    def has_migrations(self) -> bool:
        ...

    @abc.abstractmethod
    def acquire_migration_model_by_version(
        self,
        migration_version: int,
    ) -> typing.Optional[domain_migration.MigrationReadModel]:
        ...

    @abc.abstractmethod
    def acquire_migration_model_by_flag(
        self, is_applied: bool
    ) -> typing.Optional[domain_migration.MigrationReadModel]:
        ...

    @abc.abstractmethod
    def acquire_all_migration_models(
        self,
        *,
        ascending_id: bool = True,
    ) -> typing.Iterator[domain_migration.MigrationReadModel]:
        ...

    @abc.abstractmethod
    def acquire_migration_models_by_flag(
        self, *, is_applied: bool
    ) -> typing.Iterator[domain_migration.MigrationReadModel]:
        ...

    @abc.abstractmethod
    def append_migration(self, migration: domain_migration.Migration, /) -> int:
        ...

    @abc.abstractmethod
    def remove_migration(self, migration_version: int, /) -> int:
        ...

    @abc.abstractmethod
    def set_applied_flag(self, migration: domain_migration.Migration, is_applied: bool) -> int:
        ...



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = (
    "sync_scripts_with_repository",
    "recalculate_migrations_checksum",
    "raise_if_migrations_checksum_mismatch",
)

import logging
import typing

from mongorunway.application.services import migration_service
from mongorunway.domain import migration_event as domain_event
from mongorunway.domain import migration_exception as domain_exception

_LOGGER: typing.Final[logging.Logger] = logging.getLogger("root.event_handlers")


def sync_scripts_with_repository(event: domain_event.ApplicationEvent) -> None:
    service = migration_service.MigrationService(event.application.session)
    for migration in service.get_migrations():
        if not event.application.session.has_migration(migration):
            event.application.session.append_migration(migration)

            _LOGGER.info(
                "%s: migration '%s' with version %s was synced"
                " "
                "and successfully append to pending.",
                sync_scripts_with_repository.__name__,
                migration.name,
                migration.version,
            )


def recalculate_migrations_checksum(event: domain_event.ApplicationEvent) -> None:
    service = migration_service.MigrationService(event.application.session)

    for migration in event.application.session.get_all_migration_models():
        current_migration_state = service.get_migration(migration.name, migration.version)

        if current_migration_state.checksum != migration.checksum:
            event.application.session.remove_migration(migration.version)
            event.application.session.append_migration(current_migration_state)

            _LOGGER.info(
                "%s: migration file '%s' with version %s is changed, checksum successfully"
                " "
                "recalculated (%s) -> (%s).",
                recalculate_migrations_checksum.__name__,
                migration.name,
                migration.version,
                migration.checksum,
                current_migration_state.checksum,
            )


def raise_if_migrations_checksum_mismatch(event: domain_event.ApplicationEvent) -> None:
    service = migration_service.MigrationService(event.application.session)

    for migration in event.application.session.get_all_migration_models():
        current_migration_state = service.get_migration(migration.name, migration.version)
        if current_migration_state.checksum != migration.checksum:
            _LOGGER.error(
                "%s: migration file '%s' with version %s is changed, raising...",
                raise_if_migrations_checksum_mismatch.__name__,
                migration.name,
                migration.version,
            )
            raise domain_exception.MigrationFileChangedError(
                migration_name=migration.name,
                migration_version=migration.version,
            )



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = (
    "validate_migration_process",
    "validate_rule_dependencies_recursive",
)

import logging
import typing

from mongorunway import mongo
from mongorunway.domain import migration as domain_migration
from mongorunway.domain import migration_business_rule as domain_rule
from mongorunway.domain import migration_exception as domain_exception

if typing.TYPE_CHECKING:
    from mongorunway.domain import migration_context as domain_context

_LOGGER: typing.Final[logging.Logger] = logging.getLogger("mongorunway.validation_service")


def validate_rule_dependencies_recursive(
    depends_on: typing.Sequence[domain_rule.MigrationBusinessRule],
    ctx: domain_context.MigrationContext,
) -> None:
    for rule in depends_on:
        if rule.check_is_broken(ctx):
            _LOGGER.error("%s rule failed, raising...", rule.name)
            raise domain_exception.MigrationBusinessRuleBrokenError(rule)

        _LOGGER.info("%s rule successfully passed.", rule.name)

        if rule.is_independent():
            continue

        validate_rule_dependencies_recursive(
            ctx=ctx,
            depends_on=rule.depends_on,
        )


def validate_migration_process(
    migration_process: domain_migration.MigrationProcess,
    ctx: domain_context.MigrationContext,
) -> None:
    if migration_process.has_rules():
        _LOGGER.info(
            "Starting validation of migration process with version %s...",
            migration_process.migration_version,
        )

        for rule in migration_process.rules:
            if not rule.is_independent():
                validate_rule_dependencies_recursive(
                    ctx=ctx,
                    depends_on=rule.depends_on,
                )

            if rule.check_is_broken(ctx):
                _LOGGER.error("%s rule failed, raising...", rule.name)
                raise domain_exception.MigrationBusinessRuleBrokenError(rule)

            _LOGGER.info("%s rule successfully passed.", rule.name)



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = (
    "MigrationAuditlogEntry",
)

import dataclasses
import datetime
import typing
import zoneinfo

import bson.binary

if typing.TYPE_CHECKING:
    from mongorunway.domain import migration as domain_migration

_SelfT = typing.TypeVar("_SelfT", bound="MigrationAuditlogEntry")


@dataclasses.dataclass
class MigrationAuditlogEntry:
    session_id: bson.Binary
    transaction_name: str
    migration_read_model: domain_migration.MigrationReadModel
    date_fmt: str
    date: datetime.datetime = dataclasses.field(default_factory=datetime.datetime.utcnow)
    exc_name: typing.Optional[str] = None
    exc_message: typing.Optional[str] = None

    @classmethod
    def from_dict(
        cls, mapping: typing.MutableMapping[str, typing.Any], /
    ) -> MigrationAuditlogEntry:
        mapping.pop("_id", None)  # For mongo records
        return cls(**mapping)

    def is_failed(self) -> bool:
        return self.exc_name is not None or self.exc_message is not None

    def with_error(self: _SelfT, exc: BaseException, /) -> _SelfT:
        self.exc_name = type(exc).__name__
        self.exc_message = str(exc)

        return self

    def with_timezone(self: _SelfT, timezone: str) -> _SelfT:
        if timezone != "UTC":
            # Default time is utc
            try:
                self.date = self.date.astimezone(zoneinfo.ZoneInfo(timezone))
            except zoneinfo.ZoneInfoNotFoundError as exc:
                raise ModuleNotFoundError(
                    "'tzdata' module must be installed to use timezones in auditlog journals."
                ) from exc

        return self

    def format_date(self) -> str:
        return self.date.strftime(self.date_fmt)



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = (
    "MigrationError",
    "MigrationFailedError",
    "MigrationTransactionFailedError",
    "NothingToUpgradeError",
    "NothingToDowngradeError",
    "MigrationHookError",
    "MigrationFileChangedError",
)

import typing

if typing.TYPE_CHECKING:
    from mongorunway.domain import migration as domain_migration
    from mongorunway.domain import migration_business_rule as domain_rule


class MigrationError(Exception):
    __slots__: typing.Sequence[str] = ()

    pass


class MigrationFailedError(MigrationError):
    __slots__: typing.Sequence[str] = ()

    pass


class MigrationBusinessRuleBrokenError(Exception):
    __slots__ = ("rule",)

    def __init__(self, rule: domain_rule.MigrationBusinessRule) -> None:
        super().__init__(rule.render_broken_rule())
        self.rule = rule


class MigrationTransactionFailedError(MigrationFailedError):
    __slots__: typing.Sequence[str] = ("failed_migration",)

    def __init__(self, migration: domain_migration.Migration, /) -> None:
        self.failed_migration = migration

        super().__init__(
            f"Migration {migration.name!r} with version {migration.version!r} is failed."
        )


class NothingToUpgradeError(MigrationFailedError):
    __slots__: typing.Sequence[str] = ()

    def __str__(self) -> str:
        return "There are currently no pending migrations."


class NothingToDowngradeError(MigrationFailedError):
    __slots__: typing.Sequence[str] = ()

    def __str__(self) -> str:
        return "There are currently no applied migrations."


class MigrationHookError(MigrationError):
    __slots__: typing.Sequence[str] = ()

    pass


class MigrationFileChangedError(MigrationHookError):
    __slots__: typing.Sequence[str] = (
        "failed_migration_name",
        "failed_migration_version",
    )

    def __init__(self, migration_name: str, migration_version: int) -> None:
        self.failed_migration_name = migration_name
        self.failed_migration_version = migration_version

        super().__init__(
            f"Migration {migration_name!r} with version {migration_version!r} is changed."
        )



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = (
    "FileSystemConfig",
    "Config",
    "ApplicationConfig",
    "VERSIONING_STARTS_FROM",
)

import typing

import attr

if typing.TYPE_CHECKING:
    from mongorunway import mongo
    from mongorunway.application.ports import auditlog_journal as auditlog_journal_port
    from mongorunway.application.ports import filename_strategy as filename_strategy_port
    from mongorunway.application.ports import repository as repository_port
    from mongorunway.domain import migration_event as domain_event

VERSIONING_STARTS_FROM: typing.Final[int] = 1


@attr.define(frozen=True, kw_only=True, repr=True)
class FileSystemConfig:
    scripts_dir: str = attr.field(repr=True, validator=attr.validators.instance_of(str))
    config_dir: str = attr.field(repr=False, validator=attr.validators.instance_of(str))
    filename_strategy: filename_strategy_port.FilenameStrategy = attr.field(repr=True)
    use_filename_strategy: bool = attr.field(
        default=True,
        converter=attr.converters.to_bool,
        repr=True,
    )


@attr.define(frozen=True, kw_only=True, repr=True, eq=True)
class ApplicationConfig:
    app_client: mongo.Client = attr.field(repr=False)
    app_database: mongo.Database = attr.field(repr=False)
    app_repository: repository_port.MigrationModelRepository = attr.field(repr=False)
    app_auditlog_journal: typing.Optional[auditlog_journal_port.AuditlogJournal] = attr.field(
        repr=False,
    )
    app_name: str = attr.field(
        validator=[attr.validators.min_len(1), attr.validators.instance_of(str)],  # type: ignore[arg-type]
        repr=True,
        eq=True,
    )
    app_timezone: str = attr.field(
        validator=attr.validators.instance_of(str),
        default="UTC",
        repr=True,
    )
    app_date_format: str = attr.field(
        default="%Y-%m-%d %H:%M:%S",
        validator=attr.validators.instance_of(str),
    )
    app_subscribed_events: typing.Mapping[
        typing.Type[domain_event.MigrationEvent],
        typing.Sequence[domain_event.EventHandlerProxyOr[domain_event.EventHandler]],
    ] = attr.field(factory=dict, repr=False)
    app_auditlog_limit: typing.Optional[int] = attr.field(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(int)),
        converter=attr.converters.optional(int),
        repr=False,
    )

    use_logging: bool = attr.field(default=True, repr=False, converter=attr.converters.to_bool)
    use_auditlog: bool = attr.field(default=False, repr=False, converter=attr.converters.to_bool)
    use_indexing: bool = attr.field(default=False, converter=attr.converters.to_bool, repr=False)
    use_schema_validation: bool = attr.field(
        default=True,
        converter=attr.converters.to_bool,
        repr=False,
    )
    raise_on_transaction_failure: bool = attr.field(
        default=True, converter=attr.converters.to_bool, repr=False
    )

    @property
    def is_auditlog_enabled(self) -> bool:
        return self.app_auditlog_journal is not None and self.use_auditlog

    @property
    def is_logged(self) -> bool:
        return self.use_logging

    @property
    def has_unique_timezone(self) -> bool:
        return self.app_timezone != "UTC"


@attr.define(frozen=True, kw_only=True, repr=True)
class Config:
    filesystem: FileSystemConfig = attr.field(repr=True)
    application: ApplicationConfig = attr.field(repr=True)
    logging_dict: typing.Dict[str, typing.Any]



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Tuple[str, ...] = (
    "make_snake_case_global_alias",
    "CreateDatabase",
    "DropDatabase",
    "CreateCollection",
    "DropCollection",
    "InsertMany",
    "InsertOne",
    "DeleteOne",
    "DeleteMany",
    "UpdateOne",
    "UpdateMany",
    "ReplaceOne",
    "CreateIndex",
    "CreateIndexes",
    "DropIndex",
    "DropIndexes",
    "RenameCollection",
    "SendCommand",
)

import collections.abc
import functools
import inspect
import typing

import pymongo
from pymongo import results

from mongorunway import mongo
from mongorunway import util
from mongorunway.domain import migration_command as domain_command

if typing.TYPE_CHECKING:
    from mongorunway.domain import migration_context as domain_context

_T = typing.TypeVar("_T")
_CommandTT = typing.TypeVar("_CommandTT", bound=typing.Type[domain_command.AnyCommand])


@typing.overload
def make_snake_case_global_alias(obj: _CommandTT) -> _CommandTT:
    ...


@typing.overload
def make_snake_case_global_alias(
    obj: typing.MutableMapping[str, typing.Any],
) -> typing.Callable[[_CommandTT], _CommandTT]:
    ...


def make_snake_case_global_alias(
    obj: typing.Union[_CommandTT, typing.MutableMapping[str, typing.Any]],
) -> typing.Union[_CommandTT, typing.Callable[[_CommandTT], _CommandTT]]:
    def decorator(
        cls: _CommandTT,
        called_without_args: bool,
    ) -> _CommandTT:
        def func(
            *args: typing.Any,
            **kwargs: typing.Any,
        ) -> typing.Any:
            cls_instance = cls(*args, **kwargs)
            return cls_instance

        func.__name__ = util.as_snake_case(cls)
        func.__doc__ = cls.__doc__
        func.__module__ = cls.__module__

        if called_without_args:
            globals().update({func.__name__: func})
        else:
            assert isinstance(obj, collections.abc.MutableMapping)  # For type checkers only

            obj.update({func.__name__: func})

        return cls

    if inspect.isclass(obj):
        return decorator(
            typing.cast(_CommandTT, obj),
            called_without_args=True,
        )

    return typing.cast(
        typing.Callable[[_CommandTT], _CommandTT],
        functools.partial(decorator, called_without_args=False),
    )


#################################
# ~~~ CLIENT-LEVEL COMMANDS ~~~ #
#################################


@make_snake_case_global_alias
class CreateDatabase(domain_command.MigrationCommand[mongo.Database]):
    __slots__: typing.Sequence[str] = (
        "args",
        "kwargs",
        "collection",
        "database",
    )

    def __init__(
        self,
        collection: str,
        database: str,
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> None:
        self.args = args
        self.kwargs = kwargs
        self.collection = collection
        self.database = database

    def execute(self, ctx: domain_context.MigrationContext) -> mongo.Database:
        database = ctx.client.get_database(self.database)
        database.create_collection(self.collection, *self.args, **self.kwargs)
        return database


@make_snake_case_global_alias
class DropDatabase(domain_command.MigrationCommand[None]):
    __slots__: typing.Sequence[str] = (
        "args",
        "kwargs",
        "database",
    )

    def __init__(self, database: str, *args: typing.Any, **kwargs: typing.Any) -> None:
        self.database = database
        self.args = args
        self.kwargs = kwargs

    def execute(self, ctx: domain_context.MigrationContext) -> None:
        ctx.client.drop_database(self.database, *self.args, **self.kwargs)
        return None


###################################
# ~~~ DATABASE-LEVEL COMMANDS ~~~ #
###################################


@make_snake_case_global_alias
class CreateCollection(domain_command.MigrationCommand[mongo.Collection]):
    __slots__: typing.Sequence[str] = (
        "args",
        "kwargs",
        "collection",
    )

    def __init__(self, collection: str, *args: typing.Any, **kwargs: typing.Any) -> None:
        self.collection = collection
        self.args = args
        self.kwargs = kwargs

    def execute(self, ctx: domain_context.MigrationContext) -> mongo.Collection:
        collection = ctx.database.create_collection(self.collection, *self.args, **self.kwargs)
        return collection


@make_snake_case_global_alias
class DropCollection(domain_command.MigrationCommand[None]):
    __slots__: typing.Sequence[str] = (
        "args",
        "kwargs",
        "collection",
    )

    def __init__(self, collection: str, *args: typing.Any, **kwargs: typing.Any) -> None:
        self.collection = collection
        self.args = args
        self.kwargs = kwargs

    def execute(self, ctx: domain_context.MigrationContext) -> None:
        ctx.database.drop_collection(self.collection, *self.args, **self.kwargs)
        return None


@make_snake_case_global_alias
class SendCommand(domain_command.MigrationCommand[typing.Any]):
    __slots__: typing.Sequence[str] = ("args", "kwargs")

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        self.args = args
        self.kwargs = kwargs

    def execute(self, ctx: domain_context.MigrationContext) -> typing.Any:
        return ctx.database.command(*self.args, **self.kwargs)


#####################################
# ~~~ COLLECTION-LEVEL COMMANDS ~~~ #
#####################################


@make_snake_case_global_alias
class BulkWrite(domain_command.MigrationCommand[results.BulkWriteResult]):
    __slots__: typing.Sequence[str] = (
        "args",
        "kwargs",
        "collection",
        "bulk_operations",
    )

    def __init__(
        self,
        collection: str,
        bulk_operations: typing.Sequence[
            typing.Union[
                pymongo.InsertOne[mongo.DocumentType],
                pymongo.ReplaceOne[mongo.DocumentType],
                pymongo.UpdateOne,
                pymongo.UpdateMany,
                pymongo.DeleteOne,
                pymongo.DeleteMany,
            ],
        ],
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> None:
        self.collection = collection
        self.bulk_operations = bulk_operations
        self.args = args
        self.kwargs = kwargs

    def execute(self, ctx: domain_context.MigrationContext) -> results.BulkWriteResult:
        collection = ctx.database.get_collection(self.collection)
        result = collection.bulk_write(self.bulk_operations, *self.args, **self.kwargs)
        return result


@make_snake_case_global_alias
class InsertOne(domain_command.MigrationCommand[results.InsertOneResult]):
    __slots__: typing.Sequence[str] = (
        "args",
        "kwargs",
        "document",
        "collection",
    )

    def __init__(
        self,
        collection: str,
        document: typing.Any,
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> None:
        self.args = args
        self.kwargs = kwargs
        self.document = document
        self.collection = collection

    def execute(self, ctx: domain_context.MigrationContext) -> results.InsertOneResult:
        collection = ctx.database.get_collection(self.collection)
        result = collection.insert_one(self.document, *self.args, **self.kwargs)
        return result


@make_snake_case_global_alias
class InsertMany(domain_command.MigrationCommand[results.InsertManyResult]):
    __slots__: typing.Sequence[str] = (
        "args",
        "kwargs",
        "documents",
        "collection",
    )

    def __init__(
        self,
        collection: str,
        documents: typing.Iterable[typing.Any],
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> None:
        self.args = args
        self.kwargs = kwargs
        self.documents = documents
        self.collection = collection

    def execute(self, ctx: domain_context.MigrationContext) -> results.InsertManyResult:
        collection = ctx.database.get_collection(self.collection)
        result = collection.insert_many(self.documents, *self.args, **self.kwargs)
        return result


@make_snake_case_global_alias
class ReplaceOne(domain_command.MigrationCommand[results.UpdateResult]):
    __slots__: typing.Sequence[str] = (
        "args",
        "kwargs",
        "collection",
        "filter",
        "replacement",
    )

    def __init__(
        self,
        collection: str,
        filter: typing.Mapping[str, typing.Any],
        replacement: typing.Mapping[str, typing.Any],
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> None:
        self.args = args
        self.kwargs = kwargs
        self.filter = filter
        self.replacement = replacement
        self.collection = collection

    def execute(self, ctx: domain_context.MigrationContext) -> results.UpdateResult:
        collection = ctx.database.get_collection(self.collection)
        result = collection.replace_one(self.filter, self.replacement, *self.args, **self.kwargs)
        return result


@make_snake_case_global_alias
class UpdateOne(domain_command.MigrationCommand[results.UpdateResult]):
    __slots__: typing.Sequence[str] = (
        "args",
        "kwargs",
        "collection",
        "filter",
        "update",
    )

    def __init__(
        self,
        collection: str,
        filter: typing.Mapping[str, typing.Any],
        update: typing.Mapping[str, typing.Any],
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> None:
        self.args = args
        self.kwargs = kwargs
        self.filter = filter
        self.update = update
        self.collection = collection

    def execute(self, ctx: domain_context.MigrationContext) -> results.UpdateResult:
        collection = ctx.database.get_collection(self.collection)
        result = collection.update_one(self.filter, self.update, *self.args, **self.kwargs)
        return result


@make_snake_case_global_alias
class UpdateMany(domain_command.MigrationCommand[results.UpdateResult]):
    __slots__: typing.Sequence[str] = (
        "args",
        "kwargs",
        "collection",
        "filter",
        "update",
    )

    def __init__(
        self,
        collection: str,
        filter: typing.Mapping[str, typing.Any],
        update: typing.Union[
            typing.Mapping[str, typing.Any], typing.Sequence[typing.Mapping[str, typing.Any]]
        ],
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> None:
        self.args = args
        self.kwargs = kwargs
        self.filter = filter
        self.update = update
        self.collection = collection

    def execute(self, ctx: domain_context.MigrationContext) -> results.UpdateResult:
        collection = ctx.database.get_collection(self.collection)
        result = collection.update_many(self.filter, self.update, *self.args, **self.kwargs)
        return result


@make_snake_case_global_alias
class DeleteOne(domain_command.MigrationCommand[results.DeleteResult]):
    __slots__: typing.Sequence[str] = (
        "args",
        "kwargs",
        "filter",
        "collection",
    )

    def __init__(
        self,
        collection: str,
        filter: typing.Mapping[str, typing.Any],
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> None:
        self.args = args
        self.kwargs = kwargs
        self.filter = filter
        self.collection = collection

    def execute(self, ctx: domain_context.MigrationContext) -> results.DeleteResult:
        collection = ctx.database.get_collection(self.collection)
        result = collection.delete_one(self.filter, *self.args, **self.kwargs)
        return result


@make_snake_case_global_alias
class DeleteMany(domain_command.MigrationCommand[results.DeleteResult]):
    __slots__: typing.Sequence[str] = (
        "args",
        "kwargs",
        "filter",
        "collection",
    )

    def __init__(
        self,
        collection: str,
        filter: typing.Mapping[str, typing.Any],
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> None:
        self.args = args
        self.kwargs = kwargs
        self.filter = filter
        self.collection = collection

    def execute(self, ctx: domain_context.MigrationContext) -> results.DeleteResult:
        collection = ctx.database.get_collection(self.collection)
        result = collection.delete_many(self.filter, *self.args, **self.kwargs)
        return result


@make_snake_case_global_alias
class CreateIndex(domain_command.MigrationCommand[str]):
    __slots__: typing.Sequence[str] = (
        "args",
        "kwargs",
        "keys",
        "collection",
    )

    def __init__(
        self,
        collection: str,
        keys: pymongo.collection._IndexKeyHint,
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> None:
        self.args = args
        self.kwargs = kwargs
        self.keys = keys
        self.collection = collection

    def execute(self, ctx: domain_context.MigrationContext) -> str:
        collection = ctx.database.get_collection(self.collection)
        result = collection.create_index(self.keys, *self.args, **self.kwargs)
        return result


@make_snake_case_global_alias
class CreateIndexes(domain_command.MigrationCommand[typing.List[str]]):
    __slots__: typing.Sequence[str] = (
        "args",
        "kwargs",
        "keys",
        "collection",
    )

    def __init__(
        self,
        collection: str,
        keys: typing.Sequence[pymongo.IndexModel],
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> None:
        self.args = args
        self.kwargs = kwargs
        self.keys = keys
        self.collection = collection

    def execute(self, ctx: domain_context.MigrationContext) -> typing.List[str]:
        collection = ctx.database.get_collection(self.collection)
        result = collection.create_indexes(self.keys, *self.args, **self.kwargs)
        return result


@make_snake_case_global_alias
class DropIndex(domain_command.MigrationCommand[None]):
    __slots__: typing.Sequence[str] = (
        "args",
        "kwargs",
        "index_or_name",
        "collection",
    )

    def __init__(
        self,
        collection: str,
        index_or_name: pymongo.collection._IndexKeyHint,
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> None:
        self.args = args
        self.kwargs = kwargs
        self.index_or_name = index_or_name
        self.collection = collection

    def execute(self, ctx: domain_context.MigrationContext) -> None:
        collection = ctx.database.get_collection(self.collection)
        collection.drop_index(self.index_or_name, *self.args, **self.kwargs)
        return None


@make_snake_case_global_alias
class DropIndexes(domain_command.MigrationCommand[None]):
    __slots__: typing.Sequence[str] = (
        "args",
        "kwargs",
        "collection",
    )

    def __init__(
        self,
        collection: str,
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> None:
        self.args = args
        self.kwargs = kwargs
        self.collection = collection

    def execute(self, ctx: domain_context.MigrationContext) -> None:
        collection = ctx.database.get_collection(self.collection)
        collection.drop_indexes(*self.args, **self.kwargs)
        return None


@make_snake_case_global_alias
class RenameCollection(domain_command.MigrationCommand[typing.MutableMapping[str, typing.Any]]):
    __slots__: typing.Sequence[str] = (
        "args",
        "kwargs",
        "new_name",
        "collection",
    )

    def __init__(
        self,
        collection: str,
        new_name: str,
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> None:
        self.args = args
        self.kwargs = kwargs
        self.new_name = new_name
        self.collection = collection

    def execute(
        self,
        ctx: domain_context.MigrationContext,
    ) -> typing.MutableMapping[str, typing.Any]:
        collection = ctx.database.get_collection(self.collection)
        result = collection.rename(self.new_name, *self.args, **self.kwargs)
        return result


__aliases__: typing.List[str] = []
for name in __all__:
    if inspect.isclass(value := globals().get(name)):
        if issubclass(value, domain_command.MigrationCommand):
            __aliases__.append(util.as_snake_case(value))

__all__ += tuple(__aliases__)



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = (
    "show_status",
    "show_version",
    "show_auditlog_entries",
)

import typing

import terminaltables  # type: ignore[import]

from mongorunway.application import output
from mongorunway.application import use_cases
from mongorunway.presentation import formatters

if typing.TYPE_CHECKING:
    from mongorunway.application import applications


def show_auditlog_entries(
    application: applications.MigrationApp,
    verbose_exc: bool,
    ascending_date: bool,
    start: typing.Optional[typing.Sequence[str]] = None,
    end: typing.Optional[typing.Sequence[str]] = None,
    limit: typing.Optional[int] = 10,
) -> None:
    entries_result = use_cases.get_auditlog_entries(
        application,
        start=formatters.format_app_date(application, start),
        end=formatters.format_app_date(application, end),
        limit=limit,
        verbose_exc=verbose_exc,
        ascending_date=ascending_date,
    )

    if entries_result is not use_cases.UseCaseFailed:
        output.print(
            terminaltables.SingleTable(
                [
                    ["Date", "Is Failed", "Transaction Type", "Migration"],
                    *(formatters.format_auditlog_entry(entry) for entry in entries_result),
                ]
            ).table
        )


def show_version(application: applications.MigrationApp, verbose: bool) -> None:
    version_result = application.session.get_current_version()

    if version_result is not use_cases.UseCaseFailed:
        presentation = f"Current applied version is {version_result}"
        if verbose:
            all_applied_migrations_len = len(list(application.session.get_all_migration_models()))
            presentation += f" " + f"({version_result} of {all_applied_migrations_len})"

        output.print_heading(output.HEADING_LEVEL_ONE, output.TOOL_HEADING_NAME)
        output.print_success(presentation)


def show_status(
    application: applications.MigrationApp,
    verbose: bool,
    verbose_exc: bool,
    pushed_depth: int = -1,
) -> None:
    status_result = use_cases.get_status(
        application=application,
        verbose_exc=verbose_exc,
        pushed_depth=pushed_depth,
    )

    if status_result is not use_cases.UseCaseFailed:
        all_pushed_successfully, pushed_depth = status_result
        if all_pushed_successfully:
            presentation = f"All migrations applied successfully in depth {pushed_depth!r}"
        else:
            presentation = f"Applying failed in depth {pushed_depth!r}"

        if verbose:
            presentation += f" " + (
                f"({application.session.get_current_version()}"
                f" "
                f"of {len(list(application.session.get_all_migration_models()))})"
            )

            output.print_heading(output.HEADING_LEVEL_ONE, output.TOOL_HEADING_NAME)
            output.print_info(presentation)



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = (
    "MigrationEventManagerImpl",
)

import collections
import heapq
import inspect
import operator
import typing

from mongorunway.domain import migration_event as domain_event
from mongorunway.domain import migration_event_manager as domain_event_manager


class MigrationEventManagerImpl(domain_event_manager.MigrationEventManager):
    __slots__: typing.Sequence[str] = ("_event_dict", "_event_cond")

    def __init__(self) -> None:
        self._event_dict: typing.DefaultDict[
            typing.Type[domain_event.MigrationEvent],
            typing.MutableSequence[domain_event.EventHandlerProxyOr[domain_event.EventHandler]],
        ] = collections.defaultdict(list)

    def subscribe_events(
        self,
        handler: domain_event.EventHandlerProxyOr[domain_event.EventHandler],
        *events: typing.Type[domain_event.MigrationEvent],
    ) -> None:
        for event in events:
            self.subscribe_event_handler(handler, event)

    def unsubscribe_events(self, *events: typing.Type[domain_event.MigrationEvent]) -> None:
        for event in events:
            self._event_dict.pop(event)

    def subscribe_event_handler(
        self,
        handler: domain_event.EventHandlerProxyOr[domain_event.EventHandler],
        event: typing.Type[domain_event.MigrationEvent],
    ) -> None:
        self._event_dict[event].append(handler)

    def unsubscribe_event_handler(
        self,
        handler: domain_event.EventHandlerProxyOr[domain_event.EventHandler],
        event: typing.Type[domain_event.MigrationEvent],
    ) -> None:
        self._event_dict[event].remove(handler)

    def get_event_handlers_for(
        self,
        event: typing.Type[domain_event.MigrationEvent],
    ) -> typing.MutableSequence[domain_event.EventHandlerProxyOr[domain_event.EventHandler]]:
        return self._event_dict[event]

    def prioritize_handler(
        self,
        handler: domain_event.EventHandler,
        event: typing.Type[domain_event.MigrationEvent],
        priority: int,
    ) -> None:
        handlers = self.get_event_handlers_for(event)

        try:
            index = handlers.index(handler)
        except ValueError:
            raise ValueError(f"Handler {handler!r} is not subscribed for {event!r}.")

        handlers.remove(handler)
        handlers.insert(
            index,
            domain_event.EventHandlerProxy(
                handler=handler,
                priority=priority,
            ),
        )

    def unprioritize_handler_proxy(
        self,
        handler_proxy: domain_event.EventHandlerProxy,
        event: typing.Type[domain_event.MigrationEvent],
    ) -> None:
        handlers = self._event_dict[event]

        try:
            index = handlers.index(handler_proxy)
        except ValueError:
            raise ValueError(f"Handler {handler_proxy!r} is not subscribed for {event!r}.")

        handlers.remove(handler_proxy)
        handlers.insert(index, handler_proxy.handler)

    def listen(
        self,
        *events: typing.Type[domain_event.MigrationEvent],
    ) -> typing.Callable[
        [domain_event.EventHandlerProxyOr[domain_event.EventHandlerT]],
        domain_event.EventHandlerProxyOr[domain_event.EventHandlerT],
    ]:
        def decorator(
            handler: domain_event.EventHandlerProxyOr[domain_event.EventHandlerT],
        ) -> domain_event.EventHandlerProxyOr[domain_event.EventHandlerT]:
            handler_func = handler
            if isinstance(handler, domain_event.EventHandlerProxy):
                handler_func = typing.cast(
                    domain_event.EventHandlerT,
                    handler.handler,
                )

            if not events:
                signature = inspect.signature(handler_func, eval_str=True)
                try:
                    parameter = signature.parameters["event"]
                    if parameter.annotation is inspect.Parameter.empty:
                        raise
                except KeyError as exc:
                    raise ValueError(
                        f"Handler missing 'event' parameter or parameter annotation."
                    ) from exc

                if typing.get_origin(parameter.annotation) is typing.Union:
                    self.subscribe_events(handler, *typing.get_args(parameter.annotation))
                    return handler

                self.subscribe_events(handler, parameter.annotation)
                return handler

            self.subscribe_events(handler, *events)
            return handler

        return decorator

    def dispatch(self, event: domain_event.MigrationEvent) -> None:
        handlers = self.get_event_handlers_for(type(event))
        prioritized_handlers = []
        unprioritized_handlers = []

        for handler in handlers:
            if isinstance(handler, domain_event.EventHandlerProxy):
                prioritized_handlers.append(handler)
            else:
                unprioritized_handlers.append(handler)

        prioritized_handlers.sort(key=operator.attrgetter("priority"))
        heapq.heapify(prioritized_handlers)

        while prioritized_handlers:
            proxy = heapq.heappop(prioritized_handlers)
            proxy.handler(event)

        for handler in unprioritized_handlers:
            handler(event)
